from utility_tool.spiritlong_logging 	import *
from utility_tool.spiritlong_format 	import *
from utility_tool.spiritlong_string 	import *
from utility_tool.spiritlong_wechat 	import *
from utility_tool.spiritlong_email 	import *
from utility_tool.spiritlong_verify	import *


import os, smtplib
from email.mime.text			import MIMEText
from email.mime.multipart		import MIMEMultipart
from email.mime.application		import MIMEApplication
from email.header			import Header


## 邮件接口类
class  SpiritLongEmail(object):
	## 初始化
	#	email_host		邮件服务器host
	#	email_port		邮件服务器端口
	#	email_user		登录用户
	#	email_password		登录密码
	# ''
	def __init__(self, email_host, email_port, email_user, email_password):
		self.email_host		= email_host
		self.email_port		= email_port
		self.email_user		= email_user
		self.email_password	= email_password

	## 发送邮件
	#	receiver	接收者，可为数组或单个邮件
	#	subject		标题
	#	content		内容
	#	attachments	附件文件地址列表
	# ''
	def send_mail(self, receiver, subject, content, attachments):
		# 邮件接收者列表
		receiver_list			= receiver if isinstance(receiver, list) else [receiver]

		# 邮件
		message				= MIMEMultipart()
		message['From']			= self.email_user
		message['To']			= ','.join(receiver_list)
		message['Subject']		= Header(subject, 'utf-8')
		message['Accept-Language']	= "zh-CN"
		message['Accept-Charset']	= "ISO-8859-1,utf-8"

		message.attach(MIMEText(content, 'html', 'utf-8'))

		# 添加附件
		for file in attachments:
			apart	= MIMEApplication(open(file, 'rb').read())
			apart.add_header('Content-Disposition', 'attachment', filename=os.path.basename(file))
			message.attach(apart)
			
		# 发送邮件
		try:
			server	= smtplib.SMTP_SSL(self.email_host)
			server.connect(self.email_host, self.email_port)
			server.login(self.email_user, self.email_password)
			server.sendmail(self.email_user, receiver_list, message.as_string())
			server.quit()

			return True
		except Exception as ex:
			print(str(ex))

			return False
		
## 调试/测试代码
if __name__ == '__main__':
	EMAIL_HOST		= "smtp.exmail.qq.com"
	EMAIL_PORT		= 465
	EMAIL_USER		= "exon@spiritlong.com"
	EMAIL_PASSWORD		= "xxxxxx"

	# 调用库
	SL_email	= SpiritLongEmail(EMAIL_HOST, EMAIL_PORT, EMAIL_USER, EMAIL_PASSWORD)

	# 发送邮件
	SL_email.send_mail("exon@spiritlong.com", "测试邮件", "测试内容", ["__init__.py"])


from	redis				import Redis
from	urllib				import parse
from	datetime			import datetime, timedelta
from	wechatpy.oauth			import WeChatOAuth
from	wechatpy.session.redisstorage	import RedisStorage
import	wechatpy
import	functools
import	random
import	string

## 微信接口类
class  SpiritLongWechat(object):
	color_primary	= '#173177'	# 主要颜色
	color_warning	= '#8B0000'	# 警告颜色

	## 初始化
	#	appid			公众号APPID
	#	secret			公众号secret
	#	redis_session_url	REDIS连接url
	#	web_base_url		网页基础路径
	# ''
	def __init__(self, appid, secret, redis_session_url, web_base_url):
		# 微信公众号对象
		self.client	= wechatpy.client.WeChatClient(
			appid,
			secret,
			session	= RedisStorage(
				Redis.from_url(redis_session_url),
				prefix	= appid
			)
		)

		# 类属性
		self.appid		= appid
		self.web_base_url	= web_base_url
	
	## jssdk配置
	#	url	待签名的url
	# ''
	def jssdk_config(self, url):
		# 被使用一次的非重复的随机数值，15位的随机树字和ascii码
		nonce_str	= ''.join(random.choice(string.ascii_letters+string.digits) for _ in range(15))
		# 当前时间
		timestamp	= int(datetime.now().timestamp())

		# 微信公众号调用微信JS接口的临时票据
		jsapi_ticket	= self.client.jsapi.get_jsapi_ticket()

		# 微信签名
		signature	= self.client.jsapi.get_jsapi_signature(
			nonce_str,
			jsapi_ticket,
			timestamp,
			url
		)

		# 返回对应的配置参数
		return {
			"url"		: url,
			"appid"		: self.appid,
			"nonceStr"	: nonce_str,
			"timestamp"	: timestamp,
			"signature"	: signature
		}
	
	## 装饰器：捕获发送模板消息异常
	#	method	方法
	# wrapper
	def template_exception(method):
		@functools.wraps(method)
		def wrapper(*args, **kwargs):
			try:
				return method(*args, **kwargs)
			except Exception as e:
				print("++++++ 微信公众号模板消息异常 ++++++", e)
				return None
		return wrapper

	## 发送模板消息
	#	openid		微信的openid
	#	template_id	发送模板的模板id
	#	data		发送的数据
	#	url		web页面链接
	#	pagepath	小程序页面路径
	# ''
	@template_exception
	def send_template(self, openid, template_id, data, url, pagepath):
		if url is not None:
			url	= self.web_base_url+url
		miniprogram	= None

		# 跳转小程序
		if pagepath is not None:
			miniprogram	= {
				"appid"		: self.appid,
				"pagepath"	: "pages/start/index?url=/"+parse.quote(pagepath)
			}

		return self.client.message.send_template(openid, template_id, data, url, miniprogram)

## 调试/测试代码
if __name__ == '__main__':
	WECHAT_APPID		= "wx"
	WECHAT_SECRET		= ""
	WECHAT_SESSION_URL	= "redis://127.0.0.1:6379/2"
	WECHAT_BASE_URL		= "https://spiritlong-exon.com"

	# 调用库
	SL_wechat		= SpiritLongWechat(WECHAT_APPID, WECHAT_SECRET, WECHAT_SESSION_URL, WECHAT_BASE_URL)

	############################## 模板消息配置函数 ###################################

	## 登录成功提醒通知
	#	openid		用户的openid
	#	username	用户名
	#	login_mode	登录模式
	#	remark		备注
	#	url		点进去的链接
	#	pagepath	
	# ''
	def login_success(openid, username, login_mode, remark="", url="", pagepath=None):
		data	= {
			"first"		: {
				"value"	: "您已成功登录系统！",
				"color"	: SL_wechat.color_primary
			},
			"keyword1"	: {
				"value"	: username,
				"color"	: SL_wechat.color_primary
			},
			"keyword2"	: {
				"value"	: login_mode,
				"color"	: SL_wechat.color_primary
			},
			"keyword3"	: {
				"value"	: datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
				"color"	: SL_wechat.color_warning
			},
			"remark"	: {
				"value"	: remark+"点击查看详情",
				"color"	: SL_wechat.color_primary
			}
		}

		template_id	= "r5X0x7oooq09uGkDdmzc832kYJkcMs-dvRDPgRoTKeE"
		
		return SL_wechat.send_template(openid, template_id, data, url, pagepath)

	login_success("", "测试", "账号")


#!/usr/bin/python3
# coding=utf-8
###################################################################
#           ____     _     _ __  __                 
#          / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#         _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
#        /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
#           /_/                              /___/  
# Copyright (c) 2024 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.  
# @author	arthuryang
# @brief	基于池的数据库驱动MySQL
#
###################################################################  

import	MySQLdb	# pip install mysqlclient
from	MySQLdb			import OperationalError 
from	dbutils.pooled_db	import PooledDB
import	threading
import	logging
import	time
import	redis
import	pickle
import	decimal
import	json
import	datetime
import	openpyxl

# 装饰器：线程同步
def synchronized(func):
	func.__lock__ = threading.Lock()

	def synced_func(*args, **kwargs):
		with func.__lock__:
			return func(*args, **kwargs)
        
	return synced_func

# 基于池的数据库类
class MySQLDatabase(object):
	# 确保此类对每个数据库只有一个对象
	instance = {}

	@synchronized
	def __new__(cls, *args, **kwargs):
		# 根据传入的_database_config中host和db来实现每个数据库只有一个对象
		key 	= args[0]["host"] + args[0]["database"]
		if key not in cls.instance.keys():
			cls.instance[key] = super().__new__(cls)
		return cls.instance[key]

	## 类初始化
	#	_database_config	数据库配置参数，这是一个字典{user, password, host, port, database}，详细参数：https://mysqlclient.readthedocs.io/user_guide.html#mysqldb
	#	use_redis		None表示不使用redis，传入一个字典{host, port, index}，都可省略以使用默认值
	def __init__(self, _database_config, use_redis=None):
		self.database_config			= _database_config
		self.database_config["blocking"]	= True	# 连接池中如果没有可用连接，阻塞等待

		# 数据库名称
		self.database_name		= _database_config['database']

		# 默认字段
		self.default_field_names	= ['ID', 'CREATE_TIMESTAMP', 'UPDATE_TIMESTAMP', 'DELETE_TIMESTAMP']

		# 数值类型
		self.digital_field_type		= ["INT", "FLOAT", "DOUBLE", "DECIMAL"]

		# 日期类型
		self.timestamp_field_type	= ["TIMESTAMP", "TIMESTAMP ON UPDATE CURRENT_TIMESTAMP", "DATE", "DATETIME"]

		# 确保属性不会被重复初始化
		if not hasattr(self, 'connection_pool'):
			# 数据库连接池
			self.connection_pool	= PooledDB(MySQLdb, **self.database_config)
		
		# 日志初始化
		self.logger	= logging.getLogger(__name__)
		if not self.logger.handlers:
			# 避免重复添加handler
			self.logger.setLevel(logging.DEBUG)
			console_handler	= logging.StreamHandler()
			console_handler.setLevel(logging.DEBUG)
			console_handler.setFormatter(logging.Formatter('%(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S'))
			self.logger.addHandler(console_handler)

		# 更新redis配置。数据库配置中'redis'保存的是一个字典{host, port, index}，指出了使用的redis主机、端口和数据库（序号）
		if use_redis is not None:
			self.redis_host		= use_redis['host'] 	if 'host'	in use_redis else 'localhost'
			self.redis_port		= use_redis['port'] 	if 'port'	in use_redis else '6379'
			self.redis_index	= use_redis['index']	if 'index'	in use_redis else '0'
			self.redis		= redis.Redis(connection_pool=redis.ConnectionPool(
							host		= self.redis_host,
							port		= self.redis_port,
							database	= self.redis_index, 
							decode_responses=False))
		else:
			self.redis	= None

	## 对于值中的特殊字符进行替换
	def replace(self, string):
		if not string:
			return string
		string	= str(string)
		string	= string.replace("\\", "\\\\")	# 反斜杠换成两个反斜杠
		string	= string.replace("'", "\\'")	# 单引号转义
		string	= string.replace('"', '\\"')	# 双引号转义
		
		return string
	
	# ------------------------ 数据库基本操作 ------------------------
	# DDL 对数据库和表的操作	
	# 	CREATE DATABASE/TABLE
	# 	DROP DATABASE/TABLE
	# 	USE 
	# 	DESC
	#	ALTER TABLE ... ADD/DROP/CHANGE/MODIFY
	# DQL 查询
	# DML 插入/修改/删除，需要提交
	# DCL 授权
	# 操作接口：
	#	query_single	查询一条记录
	#	query		查询所有记录
	#	------------ 以下皆考虑redis缓存 ---------------
	#	query_redis	获取指定表的数据，可以指定需要的字段和ID
	#	insert		插入记录；若ID已经存在，则更新该记录
	#	update		更新记录，必须指定ID
	#	delete		删除记录（将DELET_TIMESTAMP设为当前时间，并非删除该记录）
	#	undelete	恢复删除的记录（将DELETE_TIMESTAMP设为NULL）

	## 执行SQL语句
	# 其他函数中应调用此函数来执行SQL语句
	# 私有方法，避免从外部直接调用
	#	SQL_string	要执行的SQL语句
	#	need_result	查询语句返回查询结果
	#	return_ID	当need_result==False时，返回插入语句的ID，多条语句将会返回第一个新纪录的ID；当need_result==True时，返回受影响的行数
	def _execute(self, SQL_string, need_result=False, return_ID=False):
		result	= None
		try:
			t1		= time.perf_counter()
			connection	= self.connection_pool.connection()
			cursor 		= connection.cursor()

			# 执行SQL语句
			cursor.execute(SQL_string)

			if need_result:
				columns = [desc[0] for desc in cursor.description]
				result = [dict(zip(columns, row)) for row in cursor.fetchall()]
			
			if (not need_result):
				if return_ID:
					# 获取第一个新插入记录的ID
					result	= cursor.lastrowid
				else:
					result	= cursor.rowcount
				
			cursor.close()
				
			if 'autocommit' not in self.database_config or not self.database_config['autocommit']:
				# 提交
				connection.commit()
			
			return result
		except OperationalError as ex:
			# 操作失败
			self.logger.error(str(ex))
			cursor.close()
			if 'autocommit' not in self.database_config or not self.database_config['autocommit']:
				# 回滚
				connection.rollback()
						
			return None
		except Exception as ex:
			cursor.close()
			if 'autocommit' not in self.database_config or not self.database_config['autocommit']:
				# 回滚
				connection.rollback()
			#self.logger.info(SQL_string)
			self.logger.error(str(ex))
				
			return result
	
	## 查询一条记录
	def query_single(self, SQL_string, *args):
		result	= self._execute(SQL_string+" LIMIT 1", need_result=True)
		return result[0] if result else None
		
	## 查询所有记录
	def query(self, SQL_string, *args):
		return self._execute(SQL_string, need_result=True)

	# ------------------------- 缓存查询和更新 -------------------
	# 每个记录用"database {table} {ID}"来记录，是一个HASH，field:value

	## 查询redis缓存：先尝试查询redis，若ID没有缓存，则从数据库中重新读取。注意写入和修改时已经同步到redis
	#	table		表名
	#	ID_list		ID列表，None或[]表示全部ID，可以是一个ID
	#	field_list	字段列表，[]表示全部字段，可以是一个字段的字符串
	#	valid 		是否查询有效数据（DELETE_TIMESTAMP为空）
	# 返回的
	def query_redis(self, table, ID_list=None, field_list=None, valid=True):
		# table总是全大写
		table	= table.upper()

		#  --------- 处理ID列表 ---------
		# ID_list可以是整数列表，也可以是单个整数，None或者[]表示全部ID
		if not ID_list:
			# 从数据库中查询全部ID
			SQL_string	= f"SELECT ID FROM {table}"
			if valid:
				SQL_string	+= f" WHERE DELETE_TIMESTAMP IS NULL"
			result	= self.query(SQL_string)
			if result:
				ID_list = [r['ID'] for r in result]
			else:
				# 没有ID可用
				return []
		elif isinstance(ID_list, str):
			# 支持ID_list是英文逗号分隔的ID字符串
			ID_list	= [ID for ID in ID_list.split(',')]
		elif isinstance(ID_list, int):
			# 单个数值ID处理成列表
			ID_list	= [ID_list]
		elif not isinstance(ID_list, list):
			# 不是list的其他情况
			ID_list	= []
		_ID_list	= []
		for ID in ID_list:
			try:
				# 尝试转换为整数
				_ID_list.append(int(ID))
			except:
				pass
		ID_list	= _ID_list
		
		#  --------- 处理要查询的字段 ---------
		# 标记对全部字段的查询
		all_fields	= True if not field_list else False

		if field_list is None:
			field_list	= []
		elif isinstance(field_list, str):
			# 支持字段列表是英文逗号分隔的字符串
			field_list	= [f.strip() for f in field_list.split(",")]

		# ID总是被查询
		if 'ID' not in field_list:
			# ID放在最前面
			field_list.insert(0, 'ID')
		
		# DELETE_TIMESTAMP总是被查询
		if 'DELETE_TIMESTAMP' not in field_list:
			# DELETE_TIMESTAMP放最后
			field_list.append('DELETE_TIMESTAMP')
		
		# 支持redis缓存且该表已经缓存，直接查找该表的数据
		# "database {database} {table}"是redis中的散列，key是该表的ID，对象是该记录全部数据的序列化
		redis_table	= f"database {self.database_name} {table}"
		redis_result	= []			
		if self.redis:
			for ID in ID_list:
				if self.redis.hexists(redis_table, ID):
					# 该ID已经缓存
					record	= pickle.loads(self.redis.hget(redis_table, ID))
					if (valid and record['DELETE_TIMESTAMP'] is None) or (not valid):
						# 按有效性设置来放入到结果
						redis_result.append(record)
						# 从ID_list中去掉已经查到结果的ID
						ID_list.remove(ID)
						print(f"query_redis: {ID} from redis")
			if not ID_list:
				# 全部找齐，直接返回
				return redis_result
		
		# 查询数据库全部字段，查询结果将会缓存到redis
		SQL_string	= f"SELECT * FROM {table}"
		# ID列表是必须的条件
		conditions	= [f"ID IN ({','.join([str(ID) for ID in ID_list])})"]
		
		if valid:
			# 检查DELETE_TIMESTAMP
			conditions.append(f"DELETE_TIMESTAMP IS NULL")
		if conditions:
			# SQL语句加上WHERE
			SQL_string	+= f" WHERE {' AND '.join(conditions)}"
		result	= self.query(SQL_string)

		if all_fields and result:
			# 设置全部字段
			field_list	= list(result[0].keys())

		# 缓存到redis
		self.update_redis(table, records=result)
			
		# 取需要的字段
		query_result	= [{f:r[f] for f in field_list} for r in result]
		
		return query_result+redis_result
		
	## 插入记录，若ID已经存在，则更新该记录。本实现中，insert包含了update
	#	table	要插入的表
	#	records	记录的列表[{}]
	# 如果只有一条记录，则返回其ID；否则返回None
	def insert(self, table, records):
		if not records:
			self.logger.info("新增/更新0条记录")
			return None
		
		if isinstance(records, dict):
			records		= [records]

		if not isinstance(records, list):
			self.logger.info("记录要用列表")
			return None
		
		# 已有ID的列表
		SQL_string	= f"SELECT ID FROM `{table}`"
		result		= self.query(SQL_string)
		ID_all		= [int(i['ID']) for i in result]

		# 把没有ID的选出来
		records_insert	= []
		records_update	= []
		for r in records:
			if 'ID' in r and not r['ID']:
				# 有ID而为空，去掉之
				r.pop('ID')

			if 'ID' not in r:
				# 无ID
				records_insert.append(r)
			else:
				# 有ID
				if ID_all and int(r['ID']) in ID_all:
					# ID已经存在
					records_update.append(r)
				else:
					# ID不存在
					records_insert.append(r)
		result	= None

		# UPDATE		
		if records_update:
			self.update(table, records_update)		

		# INSERT
		if records_insert: 	
			# 所有出现了的字段名合并到一个列表
			field_names_set	= set()
			for r in records_insert:
				field_names_set	= field_names_set.union(set(r.keys()))
			field_names	= list(field_names_set)

			# 构建赋值的值列表：若该字段若没有，则尝试设置为NULL
			value_strings	= []
			for r in records_insert:
				# 将每个记录的值生成字符串添加到values，要进行转义处理
				values	= []
				for f in field_names:
					if f not in r or r[f] is None:
						values.append(f"NULL")
					elif isinstance(r[f], str):
						v	= r[f]
						values.append(f"'{self.replace(v)}'")
					elif isinstance(r[f], float) or isinstance(r[f], decimal.Decimal) or isinstance(r[f], int):
						# 直接转换成字符串
						v	= str(r[f])
						values.append(f"'{self.replace(v)}'")
					elif isinstance(r[f], datetime.datetime):
						# 日期时间处理为标准格式
						v	= f"{r[f]:%Y-%m-%d %H:%M:%S}"
						values.append(f"'{self.replace(v)}'")
					else:
						# 其他数据类型都按JSON来处理
						v	= json.dumps(r[f], ensure_ascii=False)
						values.append(f"'{self.replace(v)}'")
				value_strings.append(f"( {','.join(values)} )")			
			SQL_string	= f"INSERT INTO `{table}` (`{'`,`'.join(field_names)}`) VALUES {','.join(value_strings)}"

			# 执行INSERT之前先查询最大ID
			max_ID	= 0
			if self.redis:
				result	= self.query(f"SELECT MAX(ID) FROM {table}")
				max_ID	= int(result[0]['MAX(ID)'])

			try:
				result	= self._execute(SQL_string, return_ID=True)
				self.logger.info(f"表{table}新增{len(records_insert) if result else 0}条记录，last_ID={result}")
			except Exception as ex:
				self.logger.info(f"表{table}新增{len(records_insert)}条记录失败：{ex}")
				return

			# 更新redis
			self.update_redis(table, max_ID=max_ID)

		return result	

	## 更新记录
	# 	table	记录所在的表
	# 	records	一条或多条记录的列表，每个记录必须包含ID，更新字段可以不一致
	# 返回ID_list
	def update(self, table, records):
		if not records:
			self.logger.info(f"update {table}：没有需要更新的数据")
			return False
		
		if isinstance(records, dict):
			records	= [records]
			
		# 对每一条记录进行更新
		ID_list	= []
		for r in records:
			if 'ID' not in r:
				# 必须要包含ID字段
				continue
			ID	= int(r['ID'])
			
			field_value_pairs	= []
			for f in r:
				if f!='ID':
					# 注意值要用单引号包起来，且内容需要处理
					v	= 'NULL' if r[f] is None else f"'{self.replace(r[f])}'"
					field_value_pairs.append(f"`{f}`={v}")
			SQL_string = f"UPDATE {table} SET {','.join(field_value_pairs)} WHERE ID={ID}"
			n	= self._execute(SQL_string)
			if n==1:
				# UPDATE返回受影响的行数，指定了ID应该只有一行受影响，此时表示成功更新
				# 注意：如果该记录每个域的新值都与旧值相同，即没有改变，则不会返回1
				ID_list.append(ID)
				self.logger.info(f"表{table}更新ID={ID}")
		# 更新redis
		self.update_redis(table, ID_list=ID_list)

		self.logger.info(f"表{table}更新了{len(ID_list)}条记录")

		return ID_list
		
	## 删除记录（将DELET_TIMESTAMP设为当前时间，并非删除该记录）
	def delete(self, table, ID):
		SQL_string	= f"UPDATE `{table}` SET DELETE_TIMESTAMP=CURRENT_TIMESTAMP()  WHERE `ID`={ID}"
		self._execute(SQL_string)

		# 更新redis
		self.update_redis(table, ID_list=[ID])

	## 恢复删除的记录（将DELETE_TIMESTAMP设为NULL）
	def undelete(self, table, ID):
		SQL_string	= f"UPDATE `{table}` SET DELETE_TIMESTAMP=NULL  WHERE `ID`={ID}"
		self._execute(SQL_string)
		
		# 更新redis
		self.update_redis(table, ID_list=[ID])

	# ------------------------- 工具函数 -------------------------

	## 更新redis缓存
	#	records, max_ID或ID_list要至少一种
	def update_redis(self, table, records=None, max_ID=None, ID_list=None):
		if not self.redis:
			return 
		
		redis_table	= f"database {self.database_name} {table}"
		
		if records:
			try:
				self.redis.hmset(redis_table, {r['ID']:pickle.dumps(r) for r in records})
			except Exception as ex:
				self.logger.info(f"update_redis: {ex}")				
			return
		# 使用max_ID查询
		if max_ID:
			SQL_string	= f"SELECT * FROM {table} WHERE ID>{str(max_ID)}"
		elif ID_list:
			SQL_string	= f"SELECT * FROM {table} WHERE ID in ({','.join([str(ID) for ID in ID_list])})"
		else:
			return
		# 缓存到redis：以ID（int类型）为key，value是pickle序列化的结果
		# 注意redis要设置为decode_responses=False，否则会UnicodeDecodeError
		records		= self.query(SQL_string)
		self.redis.hmset(redis_table, {r['ID']:pickle.dumps(r) for r in records})			

	## 将列表或逗号分隔的字符串转换为SQL语句中的字段列表
	def fields_string(self, table_name, fields):
		if isinstance(fields, str):
			# 字符串转列表
			fields	= fields.split(',')

		if isinstance(fields, list):
			# 列表
			try:
				fields_string	= ",".join([f"`{table_name}`.`{f.strip()}`" for f in fields])
			except Exception:
				# 可能会失败，万一不是字符串列表呢
				return None
		else:
			# 本来就不是字符串或字段列表
			return None
		
		if '*' in fields_string:
			# 不支持通配符*
			return None

		return fields_string

	## excel单元格的值转换为字符串
	def cell_string(self, sheet, row, column):
		value	= sheet.cell(row, column).value
		return str(value) if value else ""

	## 字体：主要只使用了字体名称、大小、颜色和加粗。默认黑色11号Calibri，英文数字较美观
	def cell_font(self, name='Calibri', size=11, color='000000', bold=False):
		return openpyxl.styles.Font(
			name		= name,
			size		= size,
			bold		= bold,
			vertAlign	= None,
			underline	= None,
			strike		= False,
			color		= color,
		)

	## 背景填充。默认白色
	def cell_fill(self, color='FFFFFFFF'):
		return openpyxl.styles.PatternFill(
			fill_type	= 'solid',
			fgColor		= color,
		)

	## 填充excel单元格，可以指定格式
	def set_cell(self, sheet, i, j, value, number_format='General', font=None, fill=None, border=None, alignment=None):
		cell		= sheet.cell(row=i, column=j)
		cell.value	= value
	
		cell.number_format	= number_format
		cell.font		= font if font is not None else openpyxl.styles.Font(name='Calibi', size=11, color='000000')
		# 默认无填充
		cell.fill		= fill if fill is not None else openpyxl.styles.PatternFill(fill_type=None)
		# 默认左对齐
		cell.alignment		= alignment if alignment is not None else self.alignment_left
		# 默认无边界
		cell.border		= border #if border is not None else openpyxl.styles.Border()
	
	## 自动调整指定列的宽度
	#	column		列（从1开始）
	#	width		0自动
	#	max_width	最大宽度限制
	def adjust_column_width(self, sheet, column, width=0, max_width=100):
		# 自动计算最大宽度
		if width==0:
			width	= 1
			for row in range(1, sheet.max_row+1):
				# 对于中文字符和非中文字符单独计算宽度
				s	= self.cell_string(sheet, row, column)
				w	= 0
				for c in s:
					w	+= (1 if ord(c)<128 else 2)
				width	= max(width, w+1)

		# 列宽得有最大限制
		if width>max_width:
			width	= max_width

		sheet.column_dimensions[openpyxl.utils.get_column_letter(column)].width	= width
	
	## 自动调整所有列的宽度
	def adjust_all_column_width(self, sheet, max_width=100):
		for column in range(1, sheet.max_column+1):
			self.adjust_column_width(sheet, column, max_width=max_width)

	# ------------------------ 输入输出和数据库管理操作 ------------------------
	## 字段COLUMN信息转换为我们自定义的字段字典
	def convert_COLUMN_field(self, COLUMN_field):
		# 设置字段信息
		field		= self.blank_field_string()
		field['name']	= COLUMN_field['COLUMN_NAME']

		# 保存原始的类型字符串
		field['MySQL_Type']	= COLUMN_field['COLUMN_TYPE']

		type_string	= COLUMN_field['COLUMN_TYPE'].split('(')
		field['type']	= type_string[0].upper()
		
		if field['type']=="TIMESTAMP" and "on update current_timestamp()" in COLUMN_field['EXTRA']:
			field['type']	+= " ON UPDATE CURRENT_TIMESTAMP"
			
		if len(type_string)>1:
			field['length']	= type_string[1][:-1]
		
		field['null']	= True if COLUMN_field['IS_NULLABLE']=="YES" else False
		
		if COLUMN_field['COLUMN_DEFAULT']=="current_timestamp()":
			field['default']	= "CURRENT_TIMESTAMP"
		elif COLUMN_field['COLUMN_DEFAULT']=="current_timestamp()":
			field['default']	= COLUMN_field['COLUMN_DEFAULT']
			
		if COLUMN_field['COLUMN_COMMENT']!='None':
			field['comment']	= COLUMN_field['COLUMN_COMMENT']

		if COLUMN_field['COLUMN_KEY']=='UNI':
			field['unique']	= True
		
		return field

	## 获取字段信息
	def get_field(self, table, field_name):
		# 获取字段信息
		SQL_string	= f'''
			SELECT * 
			FROM information_schema.`COLUMNS` 
			WHERE `TABLE_SCHEMA`='{self.database_config['database']}' AND 
				`TABLE_NAME`='{table}' AND 
				`COLUMN_NAME`='{field_name}'
			'''
		r	= self.query_single(SQL_string)
		if r:
			return self.convert_COLUMN_field(r)
		else:
			return None
		

	## 获取表的结构
	def get_table(self, table):
		t	= self.query(f'''
			SELECT table_name, table_comment 
			FROM information_schema.TABLES 
			WHERE table_schema = '{self.database_config['database']}' AND 
				table_name='{table}'
			''')
		if len(t)<1:
			# 没找到
			return None
		
		t	= t[0]
		
		fields = self.query(f'''
			SELECT * 
			FROM information_schema.`COLUMNS` 
			WHERE `TABLE_SCHEMA`='{self.database_config['database']}' AND 
				`TABLE_NAME`='{table}'
			''')
		
		# 排序后依次加入
		t['fields']	= []
		for f in sorted(fields, key=lambda item: item['ORDINAL_POSITION']):
			t['fields'].append(self.convert_COLUMN_field(f))
		
		return t

	## 获取数据库的结构
	def get_schema(self):
		# 读取表
		tables	= self.query(f'''
			SELECT table_name, table_comment 
			FROM information_schema.TABLES 
			WHERE table_schema = '{self.database_config['database']}'
		''')
		
		if tables==None:
			# 操作失败
			return None

		
		fields = self.query(f'''
			SELECT * 
			FROM information_schema.`COLUMNS` 
			WHERE `TABLE_SCHEMA`='{self.database_config['database']}'
		''')
		
		for t in tables:
			# 选择该表字段
			fs	= [ f for f in fields if f['TABLE_NAME']==t['table_name'] ]
			# 排序后依次加入
			t['fields']	= []
			for f in sorted(fs, key=lambda item: item['ORDINAL_POSITION']):
				t['fields'].append(self.convert_COLUMN_field(f))

			# 唯一约束
			# 该表的键（除ID主键外）
			keys	= self.query(f"SELECT * FROM information_schema.`KEY_COLUMN_USAGE` WHERE `TABLE_SCHEMA`='{self.database_config['database']}' and `TABLE_NAME`='{t['table_name']}'")
			if len(keys)>1:
				# 注意ID总是主键，排除这种情况
				constraints	= {}
				for k in keys:
					if k['CONSTRAINT_NAME'] in constraints:
						constraints[k['CONSTRAINT_NAME']].append(k['COLUMN_NAME'])
					else:
						constraints[k['CONSTRAINT_NAME']]	= [k['COLUMN_NAME']]
				t['constraints']	= constraints
		# tables按表名排序
		tables.sort(key=lambda t:t['table_name'])
		return tables

	## 空的字段SQL字符串(MySQL)
	def blank_field_string(self):
		return {
			'name'			: None,		# 字段名
			'type'			: None,		# 字段类型INT/DOUBLE/DECIMAL/CHAR/TEXT/DATETIME
			'length'		: None,		# 字段类型长度DECIMAL(10,2)/CHAR(255)
			'default'		: "NULL",	# 默认值
			'null'			: False,	# 是否可以为空，默认不能必填
			'unique'		: False,	# 是否唯一键
			'comment'		: None,		# 注释
		}

	## 从字段信息字典生成字段SQL字符串
	#	field	字段信息字典（包含name,type,length,default,null,comment）
	# 返回None表示构造失败
	def generate_field_string(self, field):
		# 检查键是否存在
		if not all(k in field for k in self.blank_field_string()):
			return None

		if not field['name']:
			return None
			
		MySQL_Type	= field['type']
		if field['length']:
			# 类型可能附带长度，要加小括号附在类型后
			MySQL_Type	+= f"({field['length']})"
		elif field['type'] in ("CHAR", "VARCHAR"):
			# CHAR/VARCHAR只使用255长度
			MySQL_Type	+= "(255)"
		s	= f"`{field['name']}` {MySQL_Type}"
		
		# 排序规则
		if field['type'] in ["CHAR", "VARCHAR", "TEXT"]:
			s	+= " CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci "
		
		# default中是初值
		if field['default'] and field['default']!='NULL':
			if field['type'] in self.digital_field_type:
				s	+= f" DEFAULT {field['default']} "
			elif field['type']=='TIMESTAMP':
				s	+= f" DEFAULT current_timestamp() " if field['default']=='CURRENT_TIMESTAMP' else ""
			else:
				s	+= f" DEFAULT '{self.replace(field['default'])}' "
		
		# 是否可以为空
		if field['null']:
			s	+= " NULL "
		else:
			s	+= " NOT NULL "

		# 注释
		if field['comment']:
			s	+= f" COMMENT '{self.replace(field['comment'])}'"

		return s

	## 增加表
	#	name	表名
	#	fields	字段信息，是一个list，每个成员是一个记录该字段信息的字典（包含name,type,length,default,on_update,null,primary,auto_increment,comment）
	#	comment	表注释
	def new_table(self, name, fields, comment=None):
		fields_strings	= []
		# 增加默认的ID字段
		fields_strings.append("`ID` INT NOT NULL AUTO_INCREMENT COMMENT 'ID'")

		# 增加各字段
		for i in range(len(fields)):
			s	= self.generate_field_string(fields[i])
			if s:
				fields_strings.append(s)

		# 增加默认的时间戳字段
		fields_strings.append("`CREATE_TIMESTAMP` TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP COMMENT '创建时间戳'")
		fields_strings.append("`UPDATE_TIMESTAMP` TIMESTAMP ON UPDATE CURRENT_TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP COMMENT '更新时间戳'")
		fields_strings.append("`DELETE_TIMESTAMP` TIMESTAMP NULL DEFAULT NULL COMMENT '删除时间戳，NULL表示未删除'")
	
		fields_strings.append('PRIMARY KEY(`ID`)')
		SQL_string	= f"CREATE TABLE `{self.database_config['database']}`.`{name}` (	\
				{','.join(fields_strings)}					\
				)  CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci ENGINE=InnoDB" 
		
		# 表注释
		if comment:
			SQL_string	+=  f" COMMENT='{self.replace(comment)}'"
	
		self._execute(SQL_string)
	
	## 修改/增加字段
	#	field	要修改/增加的字段属性，只包含要修改的属性，name必须要有
	def update_field(self, table, field, first=False):
		# field中必须要有name
		if 'name' not in field:
			return None
		
		# 读取原有字段信息
		field_old	= self.get_field(table, field['name'])
		
		if field_old:
			# 该字段已经存在	
			SQL_string	= f"ALTER TABLE `{table}` CHANGE `{field['name']}` "
		else:	
			# 需要添加字段
			SQL_string	= f"ALTER TABLE `{table}` ADD "
			field_old	= self.blank_field_string()
			field_old['name']	= field['name']
			field_old['MySQL_Type']	= field['MySQL_Type']

		if 'type' in field:
			field_old['type']	= field['type']
			if 'length' in field:
				field_old['length']	= field['length']
			else:
				field_old['length']	= None
			# 实际上是使用这个来决定数据库字段类型的
			field_old['MySQL_Type']		= field['MySQL_Type']
		
		if 'default' in field:
			field_old['default']	= field['default']
		
		if 'null' in field:
			field_old['null']	= field['null']

		if 'comment' in field:
			field_old['comment']	= field['comment']
		
		SQL_string	+= self.generate_field_string(field_old)

		if first:
			# 排到首位去
			SQL_string	+= " FIRST "

		self._execute(SQL_string)

		# 唯一性要单独处理
		if 'unique' in field:
			if field_old['unique']!= field['unique']:
				if field['unique']:	
					# 增加唯一性
					self._execute(f"ALTER TABLE `{table}` ADD UNIQUE (`{field['name']}`)")
				else:	
					# 去掉唯一性
					self._execute(f"ALTER TABLE `{table}` DROP INDEX `{field['name']}`")

		# json类型修改排序规则
		if field['type'] in ["LONGTEXT", "JSON"]:
			self._execute(f"ALTER TABLE `{table}` CHANGE `{field['name']}` `{field['name']}` LONGTEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci NULL DEFAULT NULL COMMENT '{field['comment']}'")

	## 列出/删除此数据库中所有的位于DELETE_TIMESTAMP之后的字段
	# 在导入excel文件后，将会把excel内没有的字段移动到DELETE_TIMESTAMP之后，调用此函数检查无误后，再调用此函数删除	
	def obsolete_fields(self, delete=False):
		tables	= self.get_schema()
		
		result	= []
		for t in tables:
			field_names	= []
			for f in t['fields']:
				field_names.append(f['name'])
			i	= field_names.index('DELETE_TIMESTAMP')
			for k in range(i+1, len(field_names)):
				result.append(f"{t['table_name']:20}{t['fields'][k]['name']:20}{t['fields'][k]['comment']}")
		return result
	
	# ************************************************************************
	# 				EXCEL 输入输出
	# ************************************************************************
	## 从excel导入结构
	def import_excel_schema(self, filename):
		# 读取文件
		try:
			book	= openpyxl.load_workbook(filename)
			sheet	= book['数据库表结构']			
		except Exception as ex:
			self.logger.error(str(ex))
			return None

		tables	= []
		for row in range(2, sheet.max_row+1):
			if not self.cell_string(sheet, row, 1):
				# 第一列为空：表名和注释
				tables.append({
					'table_name'	: self.cell_string(sheet, row, 2),
					'table_comment'	: self.cell_string(sheet, row, 3),
					'constraints'	: self.cell_string(sheet, row, 4),
					'fields'	: [],
				})
			else:
				type, length	= self.excel_type_string[self.cell_string(sheet, row, 4)]
				default	= self.cell_string(sheet, row, 5) 
				if not default:
					default	= "NULL"
				
				MySQL_Type	= type
				if length:
					# 类型可能附带长度，要加小括号附在类型后
					MySQL_Type	+= f"({length})"
				elif type in ("CHAR", "VARCHAR"):
					# CHAR/VARCHAR只使用255长度
					MySQL_Type	+= "(255)"
				
				tables[-1]['fields'].append({
					'name'		: self.cell_string(sheet, row, 2),	# 字段名
					'type'		: type,					# 字段类型INT/DOUBLE/DECIMAL/CHAR/TEXT/DATETIME
					'length'	: length,				# 字段类型长度DECIMAL(10,2)/CHAR(255)
					'default'	: default,				# 默认值
					'null'		: self.cell_string(sheet, row, 6)=='' ,	# 是否可以为空（只要有任何值都表示不能为空）
					'unique'	: False,				# 是否唯一键
					'comment'	: self.cell_string(sheet, row, 3),	# 注释
					'MySQL_Type'	: MySQL_Type,				# MySQL字段类型字符串
				})
		
		current_tables	= self.get_schema()
		
		if current_tables==None:
			# 读取失败
			self.logger.error("读取表结构失败")
			return None
		
		# 当前数据库表名列表
		current_table_names	= []
		for t in current_tables:
			current_table_names.append(t['table_name'])
		
		# Excel中的表名列表
		table_names		= []
		for t in tables:
			table_names.append(t['table_name'])

		# 删除excel中不存在的表
		for t in current_table_names:
			if t not in table_names:
				self._execute(f"DROP TABLE {t}")
		
		# 按excel逐个表处理
		for t in tables:
			try:
				i	= current_table_names.index(t['table_name'])
			except Exception:
				i	= -1

			if i==-1:
				# 表还不存在，创建表
				self.new_table(t['table_name'], t['fields'], t['table_comment'])
				
				self.logger.info(f"创建了表{t['table_name']}({t['table_comment']})")
			else:
				# 表已经存在，需要更新字段：只增加不删除
				# Excel中没有字段将在现有字段的备注上标注待删除并移动到末尾
				# 	注意current_tables和current_table_names是顺序一致的
				self.logger.info(f"更新表{t['table_name']}")
				
				# 表中必须有四个默认字段：ID, CREATE_TIMESTAMP, UPDATE_TIMESTAMP, DELETE_TIMESTAMP
				for field_name in ['ID', 'CREATE_TIMESTAMP', 'UPDATE_TIMESTAMP', 'DELETE_TIMESTAMP']:
					result	= self.get_field(t['table_name'], field_name)
					if not result:
						# 该字段不存在，需要创建
						if field_name=='ID':
							# ID必须存在
							pass
						elif field_name=='CREATE_TIMESTAMP':
							# 增加默认的时间戳字段CREATE_TIMESTAMP
							self._execute(f"ALTER TABLE `{t['table_name']}` ADD`CREATE_TIMESTAMP` TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP COMMENT '创建时间戳'")
						elif field_name=='UPDATE_TIMESTAMP':
							# 增加默认的时间戳字段UPDATE_TIMESTAMP
							self._execute(f"ALTER TABLE `{t['table_name']}` ADD`UPDATE_TIMESTAMP` TIMESTAMP ON UPDATE CURRENT_TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP COMMENT '更新时间戳'")
						elif field_name=='DELETE_TIMESTAMP':
							# 增加默认的时间戳字段DELETE_TIMESTAMP
							self._execute(f"ALTER TABLE `{t['table_name']}` ADD`DELETE_TIMESTAMP` TIMESTAMP NULL DEFAULT NULL COMMENT '删除时间戳，NULL表示未删除'")							
									
				# 重新排序：把ID+t['fields']+CREATE_TIMESTAMP+UPDATE_TIMESTAMP+DELETE_TIMESTAMP倒序放到开头
				# 排序过程中会使用修改的字段定义，增加excel中新的字段，在excel中没有的字段将会被移动到末尾
				# 如果表中没有这默认三字段，将会导致数据库操作返回错误
				action	= 'MODIFY' 
				self._execute(f"ALTER TABLE `{t['table_name']}` {action} `DELETE_TIMESTAMP` TIMESTAMP NULL DEFAULT NULL COMMENT '删除时间戳，NULL表示未删除' FIRST")
				self._execute(f"ALTER TABLE `{t['table_name']}` {action} `UPDATE_TIMESTAMP` TIMESTAMP ON UPDATE CURRENT_TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP COMMENT '更新时间戳' FIRST")
				self._execute(f"ALTER TABLE `{t['table_name']}` {action} `CREATE_TIMESTAMP` TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP COMMENT '创建时间戳' FIRST")
				for i in range(len(t['fields'])-1, -1, -1):
					f	= t['fields'][i]
					self.update_field(t['table_name'], f, first=True)
				self._execute(f"ALTER TABLE `{t['table_name']}` MODIFY `ID` INT(11) AUTO_INCREMENT FIRST")

				# 查询约束
				constraints	= self.query(f"SELECT * FROM information_schema.`TABLE_CONSTRAINTS` \
					WHERE `TABLE_SCHEMA`='{self.database_config['database']}' \
						AND `TABLE_NAME`='{t['table_name']}'\
						AND `CONSTRAINT_TYPE`='UNIQUE'	")
				# 去掉约束
				for c in constraints:
					self._execute(f"ALTER TABLE `{t['table_name']}` DROP INDEX `{c['CONSTRAINT_NAME']}`")
				# 更新表的注释
				self._execute(f"ALTER TABLE `{t['table_name']}`  COMMENT '{t['table_comment']}'")
			# 更新约束
			if t['constraints']!='':
				# 分号划分
				constraints	= t['constraints'].split(';')
				for c in constraints:
					name, columns	= c.split(':')
					columns	= columns.split(',')
					self._execute(f"ALTER TABLE `{t['table_name']}` ADD UNIQUE `{name}`(`{'`,`'.join(columns)}`)")
				
	## 导出结构到excel
	def export_excel_schema(self, filename=None):
		if not filename:
			# 使用默认文件名
			filename	= f"数据库{self.database_config['database']}设计.xlsx"

		book	= openpyxl.Workbook()
		
		# SHEET 数据库表
		sheet		= book.active
		sheet.title	= '数据库表结构'
			
		# 标题
		row	= 1
		self.set_cell(sheet, row, 1, '序号',	**self.cell_style_title)	
		self.set_cell(sheet, row, 2, '字段名',	**self.cell_style_title)	
		self.set_cell(sheet, row, 3, '说明',	**self.cell_style_title)	
		self.set_cell(sheet, row, 4, '类型',	**self.cell_style_title)	
		self.set_cell(sheet, row, 5, '默认值',	**self.cell_style_title)	
		self.set_cell(sheet, row, 6, '必填',	**self.cell_style_title)	
		self.set_cell(sheet, row, 7, '备注',	**self.cell_style_title)	

		# 获取表结构
		tables	= self.get_schema()
		row	= 2
		# 设置格式验证，产生下拉菜单
		validation	= openpyxl.worksheet.datavalidation.DataValidation(
			type		= 'list',
			formula1	='"'+','.join(self.excel_type_string.keys())+'"',
			allow_blank	= False,
		)
				
		if tables:
			# 逐个表生成
			for t in tables:
				# 表名和表注释
				self.set_cell(sheet, row, 2, t['table_name'], **self.cell_style_table_name)
				self.set_cell(sheet, row, 3, t['table_comment'], **self.cell_style_table_comment)
				# 约束
				if 'constraints' in t:
					constraints	= []
					for c in t['constraints']:
						if c != 'PRIMARY':
							constraints.append(f"{c}:{','.join(t['constraints'][c])}")

					self.set_cell(sheet, row, 4, ';'.join(constraints), **self.cell_style_table_comment)
				row	+= 1

				# 各字段
				for i in range(len(t['fields'])):
					f	= t['fields'][i]
					if f['name'] not in self.default_field_names:
						self.set_cell(sheet, row, 1, i, 		**self.cell_style_ID)
						self.set_cell(sheet, row, 2, f['name'],	**self.cell_style_default)
						self.set_cell(sheet, row, 3, f['comment'],	**self.cell_style_default)
						# 需要将MySQL中的字段名转换为Excel中的统一名称
						self.set_cell(sheet, row, 4, self.MySQL_type_string[f['type']],	**self.cell_style_default)
						
						# 此单元格需要验证
						validation.add(f'D{row}')

						if f['default']!="NULL":
							self.set_cell(sheet, row, 5, f['default'],	**self.cell_style_default)
						if f['null']==False:
							self.set_cell(sheet, row, 6, '●', **self.cell_style_ID)
						row	+= 1
			# 添加验证
			sheet.add_data_validation(validation)
		else:
			self.logger.info("没有数据库表")
			return

		# 自动调整列宽
		self.adjust_all_column_width(sheet)
		
		# # SHEET 数据库表
		sheet	= book.create_sheet('数据库表', 1)
		self.set_cell(sheet, 1, 1, '表名', **self.cell_style_title)
		self.set_cell(sheet, 1, 2, '说明', **self.cell_style_title)
		for i in range(len(tables)):
			# 填写excel自动计算，但这会让速度变慢
			#self.excel_cell_fill(sheet, i+2, 1, f'=INDEX(数据库表结构!$B:$B,SMALL(IF(数据库表结构!$A:$A="",ROW(数据库表结构!$A:$A),65536),ROW(A{i+2})))')
			#self.excel_cell_fill(sheet, i+2, 2, f'=INDEX(数据库表结构!$C:$C,SMALL(IF(数据库表结构!$A:$A="",ROW(数据库表结构!$A:$A),65536),ROW(A{i+2})))')
			# 直接填写表名
			self.set_cell(sheet, i+2, 1, tables[i]['table_name'], 		**self.cell_style_table_comment)
			self.set_cell(sheet, i+2, 2, tables[i]['table_comment'],	**self.cell_style_default)
		# # 自动调整列宽
		self.adjust_all_column_width(sheet)

		# SHEET 字段类型
		sheet	= book.create_sheet('字段类型', 1)
		self.set_cell(sheet, 1, 1, '字段类型', 	**self.cell_style_title)
		self.set_cell(sheet, 1, 2, 'MySQL类型', **self.cell_style_title)
		self.set_cell(sheet, 1, 3, '备注', 	**self.cell_style_title)
		row	= 2
		for k in self.excel_type_string:
			self.set_cell(sheet, row, 1, k,					**self.cell_style_default)
			self.set_cell(sheet, row, 2, self.excel_type_string[k][0],	**self.cell_style_default)
			self.set_cell(sheet, row, 3, self.excel_type_string[k][1],	**self.cell_style_default)
			row	+= 1
	
		# # 自动调整列宽
		self.adjust_all_column_width(sheet)

		# 保存
		book.save(filename)

	## 导出数据到excel
	# 每个SHEET是一个表，SHEET名称就是表名称，第一行总是字段名称
	#	table_names	表名列表[]
	def export_excel_data(self, table_names=None, filename=None):
		if not filename:
			# 使用默认文件名
			filename	= f"数据库{self.database_config['database']}数据 {time.strftime('%Y-%m-%d %H_%M_%S')}.xlsx"

		book	= openpyxl.Workbook()
		# 新建有一个sheet，删除之
		book.remove(book.active)

		if isinstance(table_names, str):
			table_names	= [table_names]

		if table_names:
			tables	= []
			for t in table_names:
				tt	= self.get_table(t)
				if tt:
					tables.append(tt)
		else:
			# 没有指定表名就导出全部表
			tables	= self.get_schema()
		for t in tables:
			sheet		= book.create_sheet(t['table_name'])
			field_names	= []
			i		= 1
			for f in t['fields']:
				# 标题行：字段名
				self.set_cell(sheet, 1, i, f['name'], **self.cell_style_title)
				comment				= openpyxl.comments.Comment(f['comment'], author="EXON")
				comment.width			= 200
				comment.hight			= 25				
				sheet.cell(1, i).comment	= comment
				i				+= 1
				field_names.append(f['name'])
			# 查询数据
			row	= 2

			result	= self.query(f"SELECT * FROM `{t['table_name']}`")

			for r in result:
				for i in range(len(field_names)):
					value		= r[field_names[i]]
					type		= t['fields'][i]['type']
					number_format	= 'General'
					if value!=None and type in self.digital_field_type[1:]:
						# 数值类型
						if type=="INT":
							value	= int(r[field_names[i]])
						else:
							value	= float(r[field_names[i]])
					elif type in self.timestamp_field_type:
						# 时间类
						number_format	= "yyyy-mm-dd hh:mm:ss"

					if value != None and value != 'NULL':
						self.set_cell(sheet, row, i+1, value, number_format=number_format)
					
				row	+= 1
			# 自动调整列宽
			self.adjust_all_column_width(sheet)
		# 保存
		book.save(filename)	

	## 从excel导入数据
	# 	filename	excel文件名
	#	overwrite	是否覆盖已有数据，默认False，将excel中的数据添加到末尾，若ID已经存在则更新数据；True则先清除该表
	def import_excel_data(self, filename, overwrite=False):
		# 读取文件
		try:
			book	= openpyxl.load_workbook(filename)
		except Exception as ex:
			self.logger.error(str(ex))
			return None
		
		for sheet_name in book.sheetnames:
			sheet	= book[sheet_name]
			# SHEET名称必须是表名称
			table	= sheet.title

			if overwrite:
				# 清除表中数据
				self._execute(f"TRUNCATE {table}")

			# 首行必须是字段名称
			field_names	= []
			for i in range(sheet.max_column):
				field_names.append(str(sheet.cell(1, i+1).value))
			# 插入，若ID存在则将更新该记录
			records	= []
			for r in range(2, sheet.max_row+1):
				record	= {}
				for k in range(len(field_names)):
					value	= sheet.cell(r, k+1).value
					record[field_names[k]]	= value if value!=None and value!='' else None
				records.append(record)
			self.logger.info(f"导入数据{len(records)}条到表{table}")
			self.insert(table, records)
		
	## ----------------------------------- 配置和定义 -----------------------------------	
	# excel中的字段名称，值是MySQL中对应的类型和长度
	excel_type_string	= {
		"STRING"	: ["CHAR", 	"255"],
		"INT"		: ["INT",	"11"],
		"JSON"		: ["JSON", 	None],		# 在 MariaDB 的实现中， JSON 类型为 longtext 类型的别名
		"DATETIME"	: ["DATETIME", 	None],
		"FLOAT"		: ["DECIMAL", 	"10,5"],	# FLOAT实际上是10.5的DECIMAL，而不是浮点数
		"DECIMAL"	: ["DECIMAL", 	"10,2"],
		"DOUBLE"	: ["DOUBLE", 	None],
		"BLOB"		: ["LONGBLOB", 	None],		# BLOB都是LONGBLOB
		"TEXT"		: ["TEXT", 	None],		# 文本只有小于255和TEXT两种
		"TIMESTAMP"	: ["TIMESTAMP", None],
	}

	# MySQL中读取的字段名称，值是excel中的对应名称
	MySQL_type_string	= {
		"CHAR"					: "STRING",
		"DATE"					: "DATETIME",
		"DATETIME"				: "DATETIME",
		"DOUBLE"				: "DOUBLE",
		"FLOAT"					: "DECIMAL",	# FLOAT实际上是10.5的DECIMAL，而不是浮点数
		"DECIMAL"				: "DECIMAL",
		"INT"					: "INT",
		"BLOB"					: "BLOB",	# BLOB都是LONGBLOB
		"LONGBLOB"				: "BLOB",	# BLOB都是LONGBLOB
		"LONGTEXT"				: "JSON",	# 在 MariaDB 的实现中， JSON 类型为 longtext 类型的别名
		"MEDIUMBLOB"				: "BLOB",
		"TEXT"					: "TEXT",
		"TIMESTAMP"				: "TIMESTAMP",
		"TIMESTAMP ON UPDATE CURRENT_TIMESTAMP"	: "TIMESTAMP",
		"VARCHAR"				: "STRING",
		"JSON"					: "JSON",
	}

	# 单元格对齐：只考虑三种基本对齐方式
	alignment_center	= openpyxl.styles.Alignment( horizontal	= 'center',	vertical = 'center',	wrap_text	= False)
	alignment_left		= openpyxl.styles.Alignment( horizontal	= 'left',	vertical = 'center',	wrap_text	= False)
	alignment_right		= openpyxl.styles.Alignment( horizontal	= 'right',	vertical = 'center',	wrap_text	= False)

	# 默认单元格样式
	cell_style_default	= {
		'font'		: openpyxl.styles.Font(name='Calibri', color='000000', size=11),
	}

	# 标题的单元格样式
	cell_style_title	= {
		'font'		: openpyxl.styles.Font(name='Calibri', color='FFFFFF', size=11),
		'fill'		: openpyxl.styles.PatternFill(fill_type='solid', fgColor='FF4169E1'),
		'alignment'	: alignment_center, 
	}

	# 表名的单元格样式
	cell_style_table_name	= {
		'font'		: openpyxl.styles.Font(name='Calibri', color='FFFFFF', size=11),
		'fill'		: openpyxl.styles.PatternFill(fill_type='solid', fgColor='FFFC760F'),
	}

	# 表注释单元格样式
	cell_style_table_comment	= {
		'font'		: openpyxl.styles.Font(name='Calibri', color='FC760F', size=11),
	}

	# ID单元格样式
	cell_style_ID		= {
		'font'		: openpyxl.styles.Font(name='Calibri', color='4169E1', size=11)
	}
	

# 调试/测试代码
if __name__ == '__main__':
	pass



#!/usr/bin/python3
# coding=utf-8
###################################################################
#           ____     _     _ __  __                 
#          / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#         _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
#        /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
#           /_/                              /___/  
# Copyright (c) 2024 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.  
# @author	arthuryang
# @brief	字符串相关工具，包括json
#
###################################################################  

import	datetime
import	re
import	os
import	json
import	decimal

import	SpiritLong_utility

## 判断一个字符串是不是浮点数
def is_float(s):
	try:
		float(s)
		# 浮点数
		return True
	except ValueError:
		pass

## 字符串tab对齐
# 	list_data	数据
#	list_align	对齐字符串
# 	length_tab	TAB空格数
# list	
@SpiritLong_utility.export
def list_content_align_tab(list_data, list_align, length_tab=8):
        list_content    = []
        list_data_temp  = list_data
        for align in list_align:
                # 字符串分割
                list_list_data  = []
                for line in list_data_temp:
                        list_list_data.append(line.strip().split(align))

                # 最小列数
                list_length_tab = []
                list_length     = len(min(list_list_data, key=len))
                
                # 每列最大长度的tab数
                for i in range(list_length):
                        list_length_tab.append(int(len(max(list(zip(*list_list_data))[i], key=len))/length_tab+1))

                # 数据拼接
                for lines in list_list_data:
                        row     = ""
                        for i in range(list_length):
                                line            = lines[i]
                                row             = row + line
                                
                                for n in range(int(len(line)/length_tab), list_length_tab[i]):
                                        row    = row+"\t"

                                if i<list_length-1:
                                        row     = row+align
                                
                        list_content.append(row + "\n")
                list_data_temp  = list_content

        return list_content

### JSON格式化日期和Decimal
class SpiritLongJsonEncoder(json.JSONEncoder):
	def default(self, obj):
		if isinstance(obj, datetime.datetime):
			return obj.strftime('%Y-%m-%d %H:%M:%S')
		elif isinstance(obj, datetime.date):
			return obj.strftime('%Y-%m-%d')
		elif isinstance(obj, decimal.Decimal):
			return str(obj).rstrip("0").rstrip(".")
		else:
			return json.JSONEncoder.default(self, obj) 
		
## 对象转json字符串
#	object_data	数据
#	default		默认值
# string
@SpiritLong_utility.export
def object_to_json(object_data, default=""):
	try:
		return json.dumps(object_data, ensure_ascii=False, cls=SpiritLongJsonEncoder)
	except Exception as ex:
		return default
	
## json字符串转对象
#	json_string	数据
#	default		默认值
# object
@SpiritLong_utility.export
def json_to_object(json_string, default={}):
	try:
		return json.loads(json_string, strict=False)
	except Exception as ex:
		return default

## json转为bytes
#	json_data	json对象
# b''
@SpiritLong_utility.export
def json_to_bytes(json_data):
	return bytes(json_data, encoding='utf8')

# 调试/测试代码
if __name__ == '__main__':
	pass




#!/usr/bin/python3
# coding=utf-8
###################################################################
#           ____     _     _ __  __                 
#          / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#         _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
#        /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
#           /_/                              /___/  
# Copyright (c) 2024 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.  
# @author	arthuryang
# @brief	excel工具集，只支持xlsx
#
###################################################################  

import	re
import	os
import	platform
import	datetime
import	chardet
import	openpyxl
import	openpyxl.styles
import	chardet
import	csv

import	SpiritLong_excel

################################### 格式/样式 ###################################
## 字体：主要只使用了字体名称、大小、颜色和加粗。默认黑色11号Calibri，英文数字较美观
@SpiritLong_excel.export
def cell_font(name='Calibri', size=11, color='000000', bold=False):
	return openpyxl.styles.Font(
		name		= name,
		size		= size,
		bold		= bold,
		vertAlign	= None,
		underline	= None,
		strike		= False,
		color		= color,
	)

## 背景填充。默认白色
@SpiritLong_excel.export
def cell_fill(color='FFFFFFFF'):
	return openpyxl.styles.PatternFill(
		fill_type	= 'solid',
		fgColor		= color,
	)

## 单元格边界
#	sides	字符串，可以是left/right/top/bottom中的一个或多个，英文逗号分隔，None或""表示四边全有
@SpiritLong_excel.export
def cell_border(color='000000', style='thin', sides=None):
	border_sides	= {
			'left'		: None,
			'right'		: None,
			'top'		: None,
			'bottom'	: None,
	}
	side		= openpyxl.styles.Side(border_style=style, color=color)
	if not sides:
		# 默认四边都有
		sides	= 'left, right, top, bottom'
	
	sides	= sides.split(',')
	for s in sides:
		ss	= s.strip()
		if ss in border_sides:
			border_sides[ss]	= side

	return openpyxl.styles.Border(**border_sides)

# 单元格对齐：只考虑三种基本对齐方式
alignment_center	= openpyxl.styles.Alignment( horizontal	= 'center',	vertical = 'center',	wrap_text	= False)
alignment_left		= openpyxl.styles.Alignment( horizontal	= 'left',	vertical = 'center',	wrap_text	= False)
alignment_right		= openpyxl.styles.Alignment( horizontal	= 'right',	vertical = 'center',	wrap_text	= False)

# 标题样式
style_title	= {
	'font'		: SpiritLong_excel.cell_font(color='FFFFFF'),
	'fill'		: SpiritLong_excel.cell_fill(color='4169E1'),
	'border'	: None,
	'alignment'	: alignment_center,
}

################################### 操作 ###################################
## 打开xlsx文件，返回book对象或None
@SpiritLong_excel.export
def open_xlsx(filename):
	try:
		book	= openpyxl.load_workbook(filename)
		return book
	except Exception as ex:
		print(str(ex))
		return None

## 设置单元格的值和格式
@SpiritLong_excel.export
def set_cell(sheet, i, j, value, number_format='General', font=None, fill=None, border=None, alignment=None):
	cell		= sheet.cell(row=i, column=j)

	# 如果是超过10位的数值长字符串，则强制设为文本类型
	if isinstance(value, str) and len(value)>=10:
		number_format	= '@'

	if isinstance(value, datetime.datetime):
		# 日期时间类需要设置格式默认值
		number_format	= 'yyyy-mm-dd hh:mm:ss' if number_format=='General' else number_format

	cell.value		= value
	cell.number_format	= number_format
	cell.font		= font if font is not None else SpiritLong_excel.cell_font()
	# 默认无填充
	cell.fill		= fill if fill is not None else openpyxl.styles.PatternFill(fill_type	= None)
	# 默认左对齐
	cell.alignment		= alignment if alignment is not None else alignment_left
	# 默认无边界
	cell.border		= border if border is not None else openpyxl.styles.Border()

## 读取带标题行的数据，每一行是一个记录，可以指定标题行位置
#	sheet		表格sheet对象
#	title_row	标题行所在行，从1开始
# 返回[{}]，列表成员是一个记录，每个记录是一个字典
@SpiritLong_excel.export
def get_records_with_title(sheet, title_row=1):
	records	= []
	titles	= []
	# 获得标题
	for column in range(1, sheet.max_column+1):
		v	= sheet.cell(title_row, column).value
		# None用空白字符串表示
		v	= v if v is not None else ""
		titles.append(v)
	
	# 获取数据
	for row in range(title_row+1, sheet.max_row+1):
		record	= {}
		for column in range(1, sheet.max_column+1):
			v				= sheet.cell(row, column).value
			# None用空白字符串表示
			v				= v if v is not None else ""
			record[titles[column-1]]	= v
		records.append(record)
	
	return records

## 自动调整指定列的宽度
#	column		列（从1开始）
#	width		0自动
#	max_width	最大宽度限制
@SpiritLong_excel.export
def adjust_column_width(sheet, column, width=0, max_width=100):
	# 自动计算最大宽度
	if width==0:
		width	= 1
		for row in range(1, sheet.max_row+1):
			# 对于中文字符和非中文字符单独计算宽度
			s	= sheet.cell(row, column).value
			s	= str(s) if s is not None else ""
			w	= 0
			for c in s:
				w	+= (1 if ord(c)<128 else 2)
			width	= max(width, w+1)

	# 列宽得有最大限制
	if width>max_width:
		width	= max_width

	sheet.column_dimensions[openpyxl.utils.get_column_letter(column)].width	= width

## 自动调整所有列的宽度
@SpiritLong_excel.export
def adjust_all_column_width(sheet, max_width=100):
	for column in range(1, sheet.max_column+1):
		SpiritLong_excel.adjust_column_width(sheet, column, max_width=max_width)

## 将'A1'转换为（1,1)
@SpiritLong_excel.export
def cell_coordinate(cell_string):
	column_letters, row	= openpyxl.utils.cell.coordinate_from_string(cell_string)
	column			= openpyxl.utils.cell.column_index_from_string(column_letters)
	return (row, column)

## 将（1，1）转换为'A1'
@SpiritLong_excel.export
def cell_code(row, column):
	return f"{openpyxl.utils.cell.get_column_letter(column)}{row}"

## 将记录保存到sheet
#	 	sheet 		sheet对象
#	{}	sheet_data	记录数据：{titles:{key:列名}或[列名], records=[{key:value}]
# 	{}	style_title	标题样式：{font,fill,border,alignment,number_format}
# 	{}	style_text	内容样式：{font,fill,border,alignment,number_format}	
# sheets中的titles为[列名]时，表示key和列名一致。records中的key可以和列名不一致，此时必须要在titles中指定{列名:key}
# records可以为空，此时将创建一个空的文件作为模板
@SpiritLong_excel.export
def records_to_sheet(sheet, sheet_data, style_title=style_title, style_text={}):
	# 判断是否有数据
	titles_valid	= 'titles' in sheet_data and sheet_data['titles']
	data_valid	= 'records' in sheet_data and sheet_data['records']
	# 字段名
	if not titles_valid:
		# 未提供则使用数据中出现了的字段名
		if data_valid:
			titles	= {k:k for r in sheet_data['records'] for k in r}
		else:
			titles	= []
	elif isinstance(sheet_data['titles'], list):
		# []则表示字段名和标题名相同
		titles	= {t:t for t in sheet_data['titles']}
	elif isinstance(sheet_data['titles'], dict):
		# {key:列名}
		titles	= sheet_data['titles']
	else:			
		print(f"sheet_data['titles']必须为标题列表、字典或None")
		return

	# 写入标题
	for i, t in enumerate(titles.values()):
		SpiritLong_excel.set_cell(sheet, 1, i+1, t, **style_title)
	
	# 写入内容
	if data_valid:
		for i, r in enumerate(sheet_data['records']):
			for j, field in enumerate(titles):
				if field in r.keys():
					SpiritLong_excel.set_cell(sheet, i+2, j+1, r[field], **style_text)	

	# 自动调整列宽
	if data_valid :
		SpiritLong_excel.adjust_all_column_width(sheet)
		
## 保存记录到excel文件
#	str 		filename 	文件名,带路径
#	{sheet_name:{}}	data		字典中每个键对应于一个sheet，其值是{titles:{key:列名}或[列名], records:[{key:value}]
# 	{}		style_title	标题样式,{font,fill,border,alignment,number_format}
# 	{}		style_text	内容样式,{font,fill,border,alignment,number_format}	
# sheets中的titles为[列名]时，表示key和列名一致。records中的key可以和列名不一致，此时必须要在titles中指定{列名:key}
# titles可以为空，此时将按records中的数据来生成列名
# records可以为空，此时将创建一个空的文件作为模板
@SpiritLong_excel.export
def records_to_excel(filename, data, style_title=style_title, style_text={}):
	if not data:
		return
	
	book 	= openpyxl.Workbook()
	# 新建有一个sheet，删除之
	book.remove(book.active)
	
	for sheet_name in data:
		sheet	= book.create_sheet(sheet_name)
		SpiritLong_excel.records_to_sheet(sheet, data[sheet_name], style_title, style_text)
	# 保存文件，确保是xlsx后缀名
	if not filename.endswith(".xlsx"):
		filename	+= ".xlsx"
	book.save(filename)

## CSV文件转换成xlsx
@SpiritLong_excel.export
def csv_save_as_xlsx(csv_file, xlsx_file=None, overwrite=False):
	filename, extesion	= os.path.splitext(csv_file)
	if extesion!='.csv':
		return
	if xlsx_file is None:
		xlsx_file	= filename+'.xlsx'

	if overwrite==False and os.path.exists(xlsx_file):
		# 若已经存在则不转换，除非被强制要求
		return

	# 先检查编码
	with open (csv_file,'rb') as f:
		encoding	= chardet.detect(f.read())["encoding"]
	# 确保用GBK，GB2312支持的汉字较少
	encoding	= 'GBK' if encoding=='GB2312' else encoding

	book	= openpyxl.Workbook()
	sheet	= book.active
	
	with open(csv_file, encoding=encoding) as f:
		reader	= csv.reader(f)
		for row in reader:
			sheet.append(row)
	book.save(xlsx_file)

if platform.system()=="Windows":
	import xlwings
## 将xls转换为xlsx文件
#	xls_file	要转换的xls
#	xlsx_file	若为None则和xls取相同文件名的xlsx文件
@SpiritLong_excel.export
def xls_save_as_xlsx(xls_file, xlsx_file=None, overwrite=False):
	filename, extesion	= os.path.splitext(xls_file)
	if extesion!='.xls':
		return
	if xlsx_file is None:
		xlsx_file	= filename+'.xlsx'

	if overwrite==False and os.path.exists(xlsx_file):
		# 若已经存在则不转换，除非被强制要求
		return
	
	# 注意：只能在windows下运行
	if platform.system()!="Windows":
		print("此功能只能在windows平台使用")
	else:
		xlwings_app = xlwings.App(visible=False, add_book=False)
		work_book = xlwings_app.books.open(xls_file, mode='r')
		work_book.save(xlsx_file)
		work_book.close()
		xlwings_app.quit()

		print(f"{xls_file}已经转换并保存到：{xlsx_file}")
	return
	
# 调试/测试代码
if __name__ == '__main__':
	# data = {
	# 	"1" : {
	# 		'titles' : { 'key1' : '名称'},
	# 		'records' : [
	# 			{'key1' : 32432 },
	# 			{'key1' : "1920-1-1" },
	# 			{'key1' : "1920-1-1" },
	# 			{'key1' : "12345678901" },
	# 			{'key1' : "510215198301011234" },
	# 			{'key1' : 1.256 },
	# 			{'key1' : "1920-1-1" },
	# 			{'key1' : "1920-1-1" },
	# 		],
	# 	}
	# }
	# #SpiritLong_excel.records_to_excel("t.xlsx", data)

	# book	= openpyxl.Workbook()
	# sheet	= book.active
	# SpiritLong_excel.set_cell(sheet, 1, 1, "321", border=cell_border())
	# SpiritLong_excel.set_cell(sheet, 2, 3, data["1"]['records'][3]['key1'], border=cell_border(sides='left, bottom'))
	# book.save('t.xlsx')
	# SpiritLong_excel.csv_save_as_xlsx('1.csv', overwrite=True)
	pass


###
#******************************************************************
# *           ____     _     _ __  __                 
# *          / __/__  (_)___(_) /_/ /  ___  ___  ___ _
# *         _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
# *        /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
# *           /_/                              /___/  
# * Copyright (c) 2023 Chongqing Spiritlong Technology Co., Ltd.
# * All rights reserved.  
# * @author	shun
# * @brief	格式化输出工具函数
# *
# *****************************************************************/

import  json
import	time
import  base64
from    datetime	import date
from    datetime	import datetime
from    decimal 	import Decimal
from    json 	        import JSONEncoder

### JSON格式化日期和Decimal
class SpiritLongJsonEncoder(json.JSONEncoder):
	def default(self, obj):
		if isinstance(obj, datetime):
			return obj.strftime('%Y-%m-%d %H:%M:%S')
		elif isinstance(obj, date):
			return obj.strftime('%Y-%m-%d')
		elif isinstance(obj, Decimal):
			return str(obj).rstrip("0").rstrip(".")
		else:
			return JSONEncoder.default(self, obj) 

## 格式化输出
#	code		代码
#	message		描述
#	data		数据
# {}
def format_error(code, message, data=None):
	return {
		"code"		: code,
		"message"	: message,
		"data"		: data
	}

## json转为bytes
#	json_data	json对象
# b''
def json_to_bytes(json_data):
	return bytes(json_data, encoding='utf8')

## json转字符串
#	json_data	json对象
# ''
def json_to_string(json_data):
	# 中文不转义
	return json.dumps(json_data, ensure_ascii=False)

## 字符串转json
#	json_data	json字符串
# object
def string_to_json(json_string):
	return json.loads(json_string)

## 格式化数据-成功
#	data		数据
# b''
def success(data=None):
	error		= format_error(0, '请求成功')
	error['data']	= data
	
	return json_to_bytes(json.dumps(error, ensure_ascii=False, cls=SpiritLongJsonEncoder))

## 格式化数据-失败
#	code		代码
#	message		描述
#	data		数据
# b''
def fail(message=None, code=500, data=None):
	# 0	请求成功
	# 400	TOKEN错误
	# 401	API不存在
	# 402	参数不存在
	# 500	自定义错误
	error	= format_error(500, '自定义错误')
	if code!=500:
		error['code']		= code
	
	if message is not None:
		error['message']	= message

	if data is not None:
		error['data']		= data
	
	return json_to_bytes(json.dumps(error))

## 字符串是否为空
#	data		数据
# bool
def is_none(data):
	if data==None or data=="None" or data=="" or len(str(data))==0:
		return True
	return False

## base64数据写入文件
#	data		base64格式数据
# 	file		文件名称，带路径
#
def base64_to_file(file, data):
	with open(file, 'wb') as f:
		f.write(base64.b64decode(data))

## 数组转元组格式字符串
#	list_data	数据
# 	is_number	内容是否是数字
# str
def list_to_touple_string(list_data, is_number=False):
	if is_number:
		return "(" + ','.join(str(i) for i in list_data) + ")"
	else:
		return "(" + ','.join('"'+str(i)+'"' for i in list_data) + ")"

## 数组字典提取指定KEY数据组成数组
#	list_dict	数据
# 	key		字典的key
# []
def list_dict_to_list(list_dict, key="ID"):
	if len(list_dict)<1 or key not in list_dict[0].keys():
		return []
	
	return [item[key] for item in list_dict]

## 数组字典提取指定KEY数据组成字典
#	list_dict	数据
# 	key_key		新字典的key在数组字典中的key
# 	key_value	新字典的value在数组字典中的key
# {}
def list_dict_to_dict(list_dict, key_key="ID", key_value="NAME"):
	if len(list_dict)<1 or key_key not in list_dict[0].keys():
		return {}
	
	return {item[key_key]:item[key_value] for item in list_dict}

## 获取浮点数
# 	data_string	数据
# float	
def format_float(data_string):
	try:
		return float(data_string)
	except Exception as ex:
		return 0
	
## 获取整数
# 	data_string	数据
# int	
def format_int(data_string):
	try:
		return int(data_string)
	except Exception as ex:
		return 0
	
## 获取日期和时间
# 	data_string	数据
# datetime
def format_datetime(data_string):
	try:
		return datetime.strptime(data_string, '%Y-%m-%d %H:%M:%S')
	except Exception as ex:
		return None
	
# 获取日期	
#  	data_string	数据
# datetime	
def format_date(data_string):
	try:
		return datetime.strptime(data_string,'%Y-%m-%d').date()
	except Exception as ex:
		return None

## 列表转元组字符串	如果列表长度大于1则返回元组，等于1则返回字符串
#	list_data	列表数据
# tuple/string
def list_to_tuple_string(list_data):
	if len(list_data) > 1:
		return tuple(list_data)
	if len(list_data)==1:
		if isinstance(list_data[0], (int, float)):
			return f'''({list_data[0]})'''
		else:
			return f'''("{list_data[0]}")'''
	return None

## 对象转json字符串
#	object_data	数据
#	default		默认值
# string
def object_to_json(object_data, default=""):
	try:
		return json.dumps(object_data, ensure_ascii=False, cls=SpiritLongJsonEncoder)
	except Exception as ex:
		return default
	

## json字符串转对象
#	json_string	数据
#	default		默认值
# object
def json_to_object(json_string, default={}):
	try:
		return json.loads(json_string, strict=False)
	except Exception as ex:
		return default

## 获取年月日
#	None
# 20230907
def get_string_date():
	return time.strftime('%Y%m%d',time.localtime(time.time()))

## 获取年月日时分秒
#	None
# 20230907121212
def get_string_datetime():
	return time.strftime('%Y%m%d%H%M%S',time.localtime(time.time()))

## 获取时分秒
#	None
# 12:12:12
def get_string_time():
	return time.strftime('%H:%M:%S',time.localtime(time.time()))

## 获取年月日时分秒和数据库时间一致
#	None
# 2023-09-07 12:12:12
def get_string_datetime_timestamp():
	return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time.time()))

## 获取年月日短横线相连
#	None
# 2023-09-07
def get_string_date_timestamp():
	return time.strftime('%Y-%m-%d', time.localtime(time.time()))


## 字符串tab对齐
# 	list_data	数据
#	list_align	对齐字符串
# 	length_tab	TAB空格数
# list	
def list_content_align_tab(list_data, list_align, length_tab=8):
        list_content    = []
        list_data_temp  = list_data
        for align in list_align:

                # 字符串分割
                list_list_data  = []
                for line in list_data_temp:
                        list_list_data.append(line.strip().split(align))

                # 最小列数
                list_length_tab = []
                list_length     = len(min(list_list_data, key=len))
                
                # 每列最大长度的tab数
                for i in range(list_length):
                        list_length_tab.append(int(len(max(list(zip(*list_list_data))[i], key=len))/length_tab+1))

                # 数据拼接
                for lines in list_list_data:
                        row     = ""
                        for i in range(list_length):
                                line            = lines[i]
                                row             = row + line
                                
                                for n in range(int(len(line)/length_tab), list_length_tab[i]):
                                        row    = row+"\t"

                                if i<list_length-1:
                                        row     = row+align
                                
                        list_content.append(row + "\n")
                list_data_temp  = list_content

        return list_content



### 调试/测试代码
if __name__ == '__main__':
	pass


#!/usr/bin/python3
# coding=utf-8
###################################################################
#           ____     _     _ __  __                 
#          / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#         _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
#        /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
#           /_/                              /___/  
# Copyright (c) 2024 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.  
# @author	arthuryang
# @brief	文件工具集
#
###################################################################  

import os

import	SpiritLong_utility

## 返回指定目录下（包括子目录）指定扩展名（列表）的文件列表
#	extensions	多个扩展名列表，空表示全部扩展名
# 返回的列表中，每个元素是一个文件名和扩展名的列表，以便后继处理
@SpiritLong_utility.export
def walk_for_files(directory, extensions=[]):
	if isinstance(extensions, str):
		extensions	= [extensions]

	file_list	= []
	for home, dirs, files in os.walk(directory):
		for f in files:
			# 完整路径名称
			filename	= os.path.join(home, f)
			name, extension	= os.path.splitext(filename)
			if not extensions or extension in extensions :
				file_list.append([name, extension])
	return file_list

# 测试
if __name__ == "__main__":
	result	= walk_for_files(r"C:\Users\yangs\Documents\SVN\SVN2024_SPIRITLONG\财务\数据\2024\原始数据\代发薪酬")
	for r in result:
		print(r)



#******************************************************************
# *           ____     _     _ __  __                 
# *          / __/__  (_)___(_) /_/ /  ___  ___  ___ _
# *         _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
# *        /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
# *           /_/                              /___/  
# * Copyright (c) 2023 Chongqing Spiritlong Technology Co., Ltd.
# * All rights reserved.  
# * @author	shun
# * @brief	logging日志支持输入多个参数
# *
# *****************************************************************/

import 	logging

### logging日志支持输入多个参数
class SpiritLongLoggingFormatter(logging.Formatter):
	def format(self, record):
		# 第一个参数为数字
		if type(record.msg) is int:
			record.msg 	= str(record.msg)
		
		# dict和tuple分别处理
		if type(record.args) is dict:
			record.msg	= record.msg + ", " + str(record.args)
		elif type(record.args) is tuple:
			for item in record.args:
				record.msg	= record.msg + ", " +str(item)
		
		# 重置参数
		record.args	= ()
		
		return super().format(record)

### 调试/测试代码
if __name__ == '__main__':
	from datetime import datetime
	
	# 基础配置
	logger		= logging.getLogger()
	handler		= logging.StreamHandler()
	handler.setFormatter(SpiritLongLoggingFormatter('%(asctime)s - %(levelname)s - %(message)s'))
	logging.basicConfig(level=logging.INFO, handlers=[handler])

	# 测试
	logger.info("debug message", "test",666)
	logger.info(666)
	logger.info(666, 888)
	logger.info([{"test1":666},{"test2":666}])
	logger.info({"test":666})
	logger.info("debug message", {"test":666})
	logger.info("debug message", {"test":666}, 666)
	logger.info("debug message", [{"test1":666},{"test2":666}])


#******************************************************************
# *           ____     _     _ __  __                 
# *          / __/__  (_)___(_) /_/ /  ___  ___  ___ _
# *         _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
# *        /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
# *           /_/                              /___/  
# * Copyright (c) 2023 Chongqing Spiritlong Technology Co., Ltd.
# * All rights reserved.  
# * @author	shun
# * @brief	字符串、字符串数组处理
# *
# *****************************************************************/


# 获取字符串在的所有行
def list_content_in(list_data, list_key):
        list_content    = []

        for line in list_data:
                for content in list_key:
                        if line.find(content)>-1:
                                line    = line.strip()

                                list_content.append(line+"\n")
                                
                                # 防止行重复
                                break
        return list_content

# 获取字符串之间的行
def list_content_between(list_data, start, end, is_contain=True):
	count_line      = 0
	list_content    = []
	for line in list_data:
                # 开始
		if line.find(start)>-1:
			count_line   = 1
			if is_contain:
				count_line      = 2
                
		if count_line > 1:
                        
			# 结束
			if line.find(end)>-1:
				count_line      = 0
				list_content.append("\n")
				if is_contain==False:
					continue

			line    = line.rstrip()
			list_content.append(line+"\n")
                
		if count_line > 0:
			count_line      = count_line + 1

	return list_content



#!/usr/bin/python3
# coding=utf-8
###################################################################
#           ____     _     _ __  __                 
#          / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#         _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
#        /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
#           /_/                              /___/  
# Copyright (c) 2024 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.  
# @author	arthuryang
# @brief	数学和数值相关工具
#
###################################################################  

import	decimal
import	re
import	os

import	SpiritLong_utility

## 将浮点数或者字符串转换成两位小数的decimal类型
@SpiritLong_utility.export
def decimal2(k):
	try:
		value	= decimal.Decimal(k).quantize(decimal.Decimal('0.00'), decimal.ROUND_HALF_UP)
		return value
	except:
		return 0	

# 调试/测试代码
if __name__ == '__main__':
	pass






#!/usr/bin/python3
# coding=utf-8
###################################################################
#           ____     _     _ __  __                 
#          / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#         _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
#        /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
#           /_/                              /___/  
# Copyright (c) 2024 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.  
# @author	arthuryang
# @brief	excel工具集
#
###################################################################  

import sys

# 动态地将函数添加到导出的模块中
# 注意：使用此装饰器之后，_function中不可直接调用同样使用此装饰器的函数，而要用模块来引用！
# 例如，模块m中，a()和b()都用了此装饰器，b()若调用a()则必须用m.a()才行！
# 事实上，所有导出的名称中都不能直接调用其他导出名称，而要用模块来引用
def export(_function):
	setattr(sys.modules[globals()['__name__']], _function.__name__, _function)

import SpiritLong_excel.excel

# 不是函数的只能手动添加
from	SpiritLong_excel.excel	import	\
		alignment_center,	\
		alignment_left,		\
		alignment_right,	\
		style_title


#!/usr/bin/python3
# coding=utf-8
###################################################################
#           ____     _     _ __  __                 
#          / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#         _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
#        /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
#           /_/                              /___/  
# Copyright (c) 2024 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.  
# @author	arthuryang
# @brief	日期时间相关工具
#
###################################################################  

import	datetime
import	re
import	os

import	SpiritLong_utility

## 根据日期时间字符串返回datetime对象或None
# 注意只支持2000-01-01之后的时间
@SpiritLong_utility.export
def datetime_string_parse(datetime_string):
	# 对于传入datetime对象的情况，直接返回
	if isinstance(datetime_string, datetime.datetime):
		return datetime_string
	
	# 得要让第一次能进去，后面在成功赋值之后都进不去
	t	= datetime.datetime(1970,1,1)
	def try_parse(tt, format_string):
		try:
			if tt is None or tt<datetime.datetime(2000,1,1):
				t	= datetime.datetime.strptime(datetime_string, format_string)
			else:
				t	= tt
		except:
			t	= None
		return t
	
	t	= try_parse(t, "%Y-%m-%d %H:%M:%S")
	t	= try_parse(t, "%Y-%m-%d")
	t	= try_parse(t, "%Y-%m")
	t	= try_parse(t, "%Y%m%d")
	t	= try_parse(t, "%Y%m")
	t	= try_parse(t, "%Y/%m/%d %H:%M:%S")
	t	= try_parse(t, "%Y/%m/%d")
	t	= try_parse(t, "%Y/%m")
	t	= try_parse(t, "%Y年%m月")
	t	= try_parse(t, "%Y年%m月%d日")
	
	return t

## 输入日期时间的解析
#	date		None将返回当前日期；datetime.datetime类型；字符串则必须以yyyy-MM-dd开头
#	last_day	True则返回当月最后一天的0点0分
#	last_second	True则返回当月最后一天的23点59分59秒
# 返回字典
#		'year'			
# 		'month'			
# 		'day'			
# 		'datetime'		
# 		'date_string'		
# 		'time_string'		
# 		'datetime_string'
@SpiritLong_utility.export	
def input_date(date=None, last_day=False, last_second=False):
	# 先转换
	date	= SpiritLong_utility.datetime_string_parse(str(date))
	
	# 转换不成功直接返回
	if not date:
		return date
	
	year	= date.year
	month	= date.month
	if last_day or last_second:
		# 设置为当月最后一天的23:59:59
		t	= date.replace(day=28) + datetime.timedelta(days=4)
		day	= (t-datetime.timedelta(days=t.day)).day
		if last_second:
			date	= date.replace(day=day, hour=23, minute=59, second=59)
		else:
			date	= date.replace(day=day, hour=0, minute=0, second=0)
	else:
		day	= date.day

	return {
		'year'				: year,
		'month'				: month, 
		'day'				: day,
		'datetime'			: date,
		'date_string'			: date.strftime('%Y-%m-%d'),
		'time_string'			: date.strftime('%H-%M-%S'),
		'datetime_string'		: date.strftime('%Y-%m-%d %H:%M:%S'),
		'date_string_chinese'		: date.strftime('%Y年%m月%d日'),
	}

# 调试/测试代码
if __name__ == '__main__':
	# print('fdsfds')
	# print(SpiritLong_utility.input_date('2024-1'))
	pass



## 
# Copyright (c) 2023 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.
# 
# @author	arthuryang
# @brief	word工具集，只支持docx。
##

import	docx		# pip install python-docx
import	re

import	SpiritLong_word

## 打开docx文件
@SpiritLong_word.export
def open_docx(filename):
	try:
		doc	= docx.Document(filename)
		return doc
	except Exception as ex:
		print(str(ex))
		exit		

## 保存docx文件
@SpiritLong_word.export
def save_docx(doc, filename):
	try:
		doc.save(filename)
	except Exception as ex:
		print(f"保存失败：{filename}")
		print(str(ex))
		exit

## 替换docx中两个段落之间的内容
#	match_start	前一段落的特征字符串
#	match_end	后一段落的特征字符串
#	paragraphs	要写入的段落
@SpiritLong_word.export
def insert_paragraphs(doc, match_start, match_end, paragraphs=None):
	p_start_index	= -1
	p_end_index	= -1
	# for i in range(len(doc.paragraphs)):
	# 	if re.findall(match_start, doc.paragraphs[i].text):
	# 		p_start	= i
	# 	if re.findall(match_end, doc.paragraphs[i].text):
	# 		p_end	= i
	
	for p in doc.paragraphs:
		if re.findall(match_start, p.text):
			p_start	= p
		if re.findall(match_end, p.text) and p_start:
			# 确保p_start和p_end的先后顺序
			p_end	= p 

	if p_start and p_end:
		# 都找到了，删除
		flag	= False
		for p in doc.paragraphs:
			if flag and p._element==p_end._element:
				# 结束段落
				flag	= False
			elif p._element==p_start._element:
				# 开始段落
				flag	= True
			elif flag:
				# 该删除
				SpiritLong_word.delete_paragraph(p)
		# 在开始段落之后添加
		SpiritLong_word.add_paragraphs(p_start, paragraphs)

## 删除段落
@SpiritLong_word.export
def delete_paragraph(p):
	# element才是实质的段落数据对象，Paragraph只是指向它的对象封装
	element	= p._element
	element.getparent().remove(element)
	p._p = p._element = None

## 添加多个段落
#	p_start		在此段落之后插入多个段落
#	paragraphs	列表成员每个是一个(runs, style)元组，其中runs列表成员是一个字典
@SpiritLong_word.export
def add_paragraphs(p_start, paragraphs):
	last_p	= p_start
	for p in paragraphs:
		new_p = docx.oxml.xmlchemy.OxmlElement("w:p")
		last_p._p.addnext(new_p)
		# 注意用new_p创建新段落之后，last_p将指向该新段落
		last_p	= docx.text.paragraph.Paragraph(new_p, last_p._parent)
		
		if p:
				
			# 添加此段落的各部分
			for r in p['runs']:
				new_run	= last_p.add_run(r['text'])
				if 'font_name' in r:
					new_run.font.name	= r['font_name']
					new_run._element.rPr.rFonts.set(docx.oxml.ns.qn("w:eastAsia"), new_run.font.name)
				if 'font_size' in r:
					new_run.font.size	= r['font_size']
				if 'font_color_rgb' in r:
					new_run.font.color.rgb	= r['font_color_rgb']
				if 'bold' in r:
					new_run.font.bold	= r['bold']
				# 段落中的风格也适用于全体文字
				if 'font_name' in p:
					new_run.font.name	= p['font_name']
					new_run._element.rPr.rFonts.set(docx.oxml.ns.qn("w:eastAsia"), new_run.font.name)
				if 'font_size' in p:
					new_run.font.size	= p['font_size']
				if 'font_color_rgb' in p:
					new_run.font.color.rgb	= p['font_color_rgb']
				if 'bold' in p:
					new_run.font.bold	= p['bold']
			# 添加此段落风格
			if 'style' in p:
				last_p.style	= p['style']
			if 'alignment' in p:
				last_p.alignment				= p['alignment']
			if 'left_indent' in p:
				last_p.paragraph_format.left_indent		= p['left_indent']
			if 'right_indent' in p:
				last_p.paragraph_format.right_indent		= p['right_indent']
			if 'first_line_indent' in p:
				last_p.paragraph_format.first_line_indent	= p['first_line_indent']
			if 'space_before' in p:
				last_p.paragraph_format.space_before 		= p['space_before']
			if 'space_after' in p:
				last_p.paragraph_format.space_after  		= p['space_after']
			if 'line_spacing' in p:
				last_p.paragraph_format.line_spacing 		= p['line_spacing']
			if 'line_spacing_rule' in p:
				last_p.paragraph_format.line_spacing_rule 	= p['line_spacing_rule']
	return last_p

## 返回默认段落
@SpiritLong_word.export
def new_paragraph_data(runs=[], 
		alignment=docx.enum.text.WD_PARAGRAPH_ALIGNMENT.LEFT, 
		first_line_indent	= 0,
		line_spacing_rule	= docx.enum.text.WD_LINE_SPACING.SINGLE,
		line_spacing		= docx.shared.Pt(10)	):
	return {
		'runs'			: runs,
		'alignment'		: alignment,
		'left_indent'		: docx.shared.Cm(0),
		'right_indent'		: docx.shared.Cm(0),
		'first_line_indent'	: docx.shared.Pt(first_line_indent),
		'space_before'		: docx.shared.Pt(0),
		'space_after'		: docx.shared.Pt(0),
		'line_spacing'		: line_spacing,
		'line_spacing_rule'	: line_spacing_rule,			
	}

## 返回默认run
@SpiritLong_word.export
def new_run_data(text="", font_name="宋体", font_size=12, color=docx.shared.RGBColor(0, 0, 0), bold=False, italic=False):
	return {
		"text"			: text,
		"bold"			: bold,
		"italic"		: italic,	
		"font_name"		: font_name,
		"font_size"		: docx.shared.Pt(font_size),
		"font_color_rgb"	: color,
	}		

## 在段落之后添加若干空行（第一个空行需要手动添加）
@SpiritLong_word.export
def add_blank_paragraphs(p_start, n):
	p	= p_start

	for i in range(n):
		paragraph	= {
			'runs'	: [{'text':""}],
			'style'	: p_start.style,
		}
		p		= SpiritLong_word.add_paragraphs(p, [paragraph])
	return p

## 添加指定样式的段落，支持换行符
#	p_start			在此段落之后添加
#	style			样式名称
#	style_attributes	样式字典直接指定样式
@SpiritLong_word.export
def add_text_with_style(p_start, text, style=None, style_attributes={}):
	if not style:
		# 使用p_start的样式
		style	= p_start.style
	style_attributes.update({'style':style})
		
	# 分析text中的换行符
	paragraphs	= [dict({'runs' : [{'text':t}]}, **style_attributes) for t in text.split('\n')]
	return SpiritLong_word.add_paragraphs(p_start, paragraphs)

## 寻找指定的文字（相同样式）进行替换，这些指定文字用〖〗包含
@SpiritLong_word.export
def replace_text(doc, old_text, text):
	old_text_string	= f"〖{old_text}〗"
	for p in doc.paragraphs:
		# 在每个running中寻找
		if re.findall(old_text_string, p.text):
			# 在此段落
			start	= 0
			end	= 0
			state	= 0
			for i in range(len(p.runs)):
				if state==0:
					# 还没找到〖
					if p.runs[i].text=="〖":
						state	= 1
						start	= i
				elif state==1:
					# 已经找到〖，检查到是否匹配文本
					if old_text_string==''.join([r.text for r in p.runs[start:i+1]]):
						# 文本匹配
						p.runs[start].clear()
						p.runs[start+1].text	= str(text)
						#p.runs[start].style	= p.runs[start+1].style
						for j in range(start+2, i+1):
							p.runs[j].clear()
						return p
				elif p.runs[i+1]=="〗":
					# 重新来
					state	= 0

## 在指定的文本之后添加
@SpiritLong_word.export
def append_text(doc, match_text, text):
	p	= SpiritLong_word.find_paragraph(doc, match_text)
	p.text	= p.text.replace(match_text, match_text+text)
	return p
		
## 定位指定文本所在的段落
@SpiritLong_word.export
def find_paragraph(doc, match_text):
	for p in doc.paragraphs:
		if re.findall(match_text, p.text):
			return p

## 返回文档中的样式名称列表
@SpiritLong_word.export
def get_styles(doc):
	styles = doc.styles
	return [s.name for s in styles if s.type == docx.enum.style.WD_STYLE_TYPE.PARAGRAPH]
		
## 删除word表格中的行
#	table	表对象
#	line	行序号，从0开始
@SpiritLong_word.export
def table_remove_row(table, line):
	if line>=len(table.rows):
		return 
	row	= table.rows[line]
	row._element.getparent().remove(row._element)

## 在表格中插入行
#	table	表对象
#	row	行序号从0开始）
@SpiritLong_word.export
def table_insert_row(table, row):
	tbl		= table._tbl
	successor	= tbl.tr_lst[row]
	tr		= tbl._new_tr()
	for gridCol in tbl.tblGrid.gridCol_lst:
		tc		= tr.add_tc()
		tc.width	= gridCol.w
	successor.addprevious(tr)

	style	= {
		'sz'	: 8,		# 大概对应1.0磅的粗细
		'val'	: 'single',	# 线条
		'color'	: '#000000',
	}
	for cell in table.rows[row].cells:
		SpiritLong_word.set_cell_border(cell, top=style, bottom=style, left=style, right=style)
		# 竖直居中
		cell.vertical_alignment	= docx.enum.table.WD_ALIGN_VERTICAL.CENTER
	return table.rows[row]

## 设置单元格边框
# 参数举例
#	top={"sz": 12, "val": "single", "color": "#FF0000", "space": "0"},
#	bottom={"sz": 12, "color": "#00FF00", "val": "single"},
#	left={"sz": 24, "val": "dashed", "shadow": "true"},
#	right={"sz": 12, "val": "dashed"},
@SpiritLong_word.export
def set_cell_border(cell, **kwargs):
	tc	= cell._tc
	tcPr	= tc.get_or_add_tcPr()

	# check for tag existnace, if none found, then create one
	tcBorders = tcPr.first_child_found_in("w:tcBorders")
	if tcBorders is None:
		tcBorders	= docx.oxml.OxmlElement('w:tcBorders')
		tcPr.append(tcBorders)

	# list over all available tags
	for edge in ('left', 'top', 'right', 'bottom', 'insideH', 'insideV'):
		edge_data = kwargs.get(edge)
		if edge_data:
			tag = f'w:{edge}'
		else:
			continue

		# check for tag existnace, if none found, then create one
		element	= tcBorders.find(docx.oxml.ns.qn(tag))
		if element is None:
			element	= docx.oxml.OxmlElement(tag)
			tcBorders.append(element)

		# looks like order of attributes is important
		for key in ["sz", "val", "color", "space", "shadow"]:
			if key in edge_data:
				element.set(docx.oxml.ns.qn('w:{}'.format(key)), str(edge_data[key]))

## 设置word表格中的单元格内容，包括字体和大小
@SpiritLong_word.export
def table_cell_fill(table, row, column, content, style=None, style_attributes={}):
	# 如果超出范围，就创建
	for i in range(row+1-len(table.rows)):
		table.add_row()
	for i in range(column+1-len(table.columns)):
		table.add_column(docx.shared.Cm(1))

	cell	= table.cell(row, column)
	p_start	= cell.paragraphs[0]
	SpiritLong_word.add_text_with_style(p_start, content, style, style_attributes)
	SpiritLong_word.delete_paragraph(p_start)

## 在指定段落后创建表格
@SpiritLong_word.export
def create_table(doc, paragraph, rows, columns, style='Table Grid'):
	table	= doc.add_table(rows, columns, style)
	# 将表格附着到段落之后
	paragraph._p.addnext(table._tbl)
	return table

## 合并单元格
@SpiritLong_word.export
def merge_cell(table, row_min, row_max, column_min, column_max, text, style=None, style_attributes={}):
	SpiritLong_word.table_cell_fill(table, row_min, column_min, text, style, style_attributes)
	table.cell(row_min, column_min).merge(table.cell(row_max, column_max))
	table.cell(row_min, column_min).vertical_alignment	= docx.enum.table.WD_ALIGN_VERTICAL.CENTER

## 调整列宽
@SpiritLong_word.export
def set_table_column_width_and_vertical_center(table, width_list):
	for row in table.rows:
		for i in range(len(width_list)):
			row.cells[i].width	= docx.shared.Cm(width_list[i])	
		for i in range(len(table.columns)):
			# 都纵向居中
			row.cells[i].vertical_alignment	= docx.enum.table.WD_ALIGN_VERTICAL.CENTER
			
# 调试/测试代码
if __name__ == '__main__':
	pass

